package com.qmth.distributed.print.api;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.google.gson.reflect.TypeToken;
import com.qmth.boot.api.constant.ApiConstant;
import com.qmth.distributed.print.business.bean.dto.FinalScoreDto;
import com.qmth.distributed.print.business.bean.dto.report.*;
import com.qmth.distributed.print.business.bean.result.FinalScoreResult;
import com.qmth.distributed.print.business.bean.result.ScoreResult;
import com.qmth.distributed.print.business.bean.result.report.PaperStructResult;
import com.qmth.distributed.print.business.bean.result.report.ReportResult;
import com.qmth.distributed.print.business.entity.TCPaperStruct;
import com.qmth.distributed.print.business.entity.TRBasicInfo;
import com.qmth.distributed.print.business.service.*;
import com.qmth.teachcloud.common.annotation.OperationLogDetail;
import com.qmth.teachcloud.common.contant.SystemConstant;
import com.qmth.teachcloud.common.entity.BasicExam;
import com.qmth.teachcloud.common.entity.BasicSemester;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.log.CustomizedOperationTypeEnum;
import com.qmth.teachcloud.common.service.BasicSemesterService;
import com.qmth.teachcloud.common.util.GsonUtil;
import com.qmth.teachcloud.common.util.JacksonUtil;
import com.qmth.teachcloud.common.util.ResultUtil;
import com.qmth.teachcloud.common.util.ServletUtil;
import com.qmth.teachcloud.mark.bean.archivescore.ScoreRangeVo;
import com.qmth.teachcloud.mark.entity.MarkPaper;
import com.qmth.teachcloud.mark.service.MarkPaperService;
import com.qmth.teachcloud.mark.utils.Calculator;
import io.swagger.annotations.*;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import javax.validation.Valid;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * <p>
 * 报告基本情况表 前端控制器
 * </p>
 *
 * @author wangliang
 * @since 2024-02-18
 */
@Api(tags = "课程目标达成度-报告基本情况Controller")
@RestController
@RequestMapping(ApiConstant.DEFAULT_URI_PREFIX + SystemConstant.PREFIX_URL_COURSE_DEGREE)
public class TRBasicInfoController {

    @Resource
    PrintCommonService printCommonService;

    @Resource
    MarkPaperService markPaperService;

    @Resource
    TRBasicInfoService trBasicInfoService;

    @Resource
    BasicSemesterService basicSemesterService;

    @Resource
    BasicExamService basicExamService;

    @Resource
    TCPaperStructService tcPaperStructService;

    @Resource
    TCFinalScoreService tcFinalScoreService;

    @ApiOperation(value = "报告管理列表")
    @RequestMapping(value = "/report/list", method = RequestMethod.POST)
    @OperationLogDetail(customizedOperationType = CustomizedOperationTypeEnum.SEARCH)
    @ApiResponses({@ApiResponse(code = 200, message = "分页查询", response = ScoreResult.class)})
    public Object reportList(@ApiParam(value = "考试ID", required = true) @RequestParam Long examId,
                             @ApiParam(value = "课程编码") @RequestParam(required = false) String courseCode,
                             @ApiParam(value = "分页页码", required = true) @RequestParam @Min(SystemConstant.PAGE_NUMBER_MIN) Integer pageNumber,
                             @ApiParam(value = "分页数", required = true) @RequestParam @Min(SystemConstant.PAGE_SIZE_MIN) @Max(SystemConstant.PAGE_SIZE_MAX) Integer pageSize) {
        return ResultUtil.ok(printCommonService.scoreList(new Page<>(pageNumber, pageSize), examId, courseCode));
    }

    @ApiOperation(value = "查看报告")
    @RequestMapping(value = "/report/view", method = RequestMethod.POST)
    @OperationLogDetail(customizedOperationType = CustomizedOperationTypeEnum.SEARCH)
    @ApiResponses({@ApiResponse(code = 200, message = "查看报告", response = ReportResult.class)})
    @Transactional
    public Object reportView(@ApiParam(value = "考试id", required = true) @RequestParam Long examId,
                             @ApiParam(value = "科目编码", required = true) @RequestParam String courseCode,
                             @ApiParam(value = "试卷编号", required = true) @RequestParam String paperNumber) {
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(examId, paperNumber);
        Objects.requireNonNull(markPaper, "未找到科目信息");

        SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
        TRBasicInfo trBasicInfo = trBasicInfoService.getOne(new QueryWrapper<TRBasicInfo>().lambda().eq(TRBasicInfo::getExamId, examId).eq(TRBasicInfo::getCourseCode, courseCode).eq(TRBasicInfo::getPaperNumber, paperNumber));
        if (Objects.isNull(trBasicInfo)) {
            BasicExam basicExam = basicExamService.getById(examId);
            Objects.requireNonNull(basicExam, "未找到考试信息");

            BasicSemester basicSemester = basicSemesterService.getById(basicExam.getSemesterId());
            Objects.requireNonNull(basicSemester, "未找到学期信息");

            //TODO 有测试数据，待肖飞补充授课对象和任课老师数据
            trBasicInfo = new TRBasicInfo(examId, courseCode, markPaper.getCourseName(), paperNumber, basicSemester.getName(), "测试班级1", "测试老师1", 40, sysUser.getId());
        }
        //测试数据

        //公共信息
        ReportCommonDto reportCommonDto = new ReportCommonDto(examId, courseCode, markPaper.getCourseName(), paperNumber);
        //课程基本情况
        ReportCourseBasicInfoDto reportCourseBasicInfoDto = new ReportCourseBasicInfoDto(trBasicInfo);

        //课程目标考核分布-题目信息
        TCPaperStruct tcPaperStruct = tcPaperStructService.getOne(new QueryWrapper<TCPaperStruct>().lambda().eq(TCPaperStruct::getExamId, examId).eq(TCPaperStruct::getCourseCode, courseCode).eq(TCPaperStruct::getPaperNumber, paperNumber));
        Objects.requireNonNull(tcPaperStruct, "未找到试卷蓝图信息");
        List<PaperStructResult> paperStructResultList = GsonUtil.fromJson(tcPaperStruct.getPaperStruct(), new TypeToken<List<PaperStructResult>>() {
        }.getType());

        ReportCourseEvaluationSpreadDto reportCourseEvaluationSpreadDto = new ReportCourseEvaluationSpreadDto(paperStructResultList);

        //课程目标考核分布-分数图
        FinalScoreDto finalScoreDto = tcFinalScoreService.finalScoreScoreOverView(examId, courseCode, paperNumber);
        Objects.requireNonNull(finalScoreDto, "未找到期末成绩分数信息");

        fillScoreRange(finalScoreDto, examId, paperNumber, markPaper);
        List<ScoreRangeVo> scoreRangeVoList = finalScoreDto.getScoreRange();
        if (!CollectionUtils.isEmpty(scoreRangeVoList)) {
            Integer failCount = scoreRangeVoList.get(0).getStudentCount();//第一行数据默认为不及格
            Double failRate = scoreRangeVoList.get(0).getRate();

            scoreRangeVoList.remove(0);
            List<ReportScoreRangeViewDto> scoreRangeViewDtoList = GsonUtil.fromJson(JacksonUtil.parseJson(scoreRangeVoList), new TypeToken<List<ReportScoreRangeViewDto>>() {
            }.getType());
            ReportScoreViewDto reportScoreViewDto = new ReportScoreViewDto(failCount, failRate, scoreRangeViewDtoList);
            reportCourseEvaluationSpreadDto.setScoreList(reportScoreViewDto);
        }
        trBasicInfo.setCourseEvaluationSpread(JacksonUtil.parseJson(reportCourseEvaluationSpreadDto));

        //查询考生数据
        List<FinalScoreResult> finalScoreResultList = tcFinalScoreService.examStudentOverview(examId, courseCode, paperNumber);
        Map<String, List<Double>> usualScoreMap = new LinkedHashMap<>();//汇总平常成绩分数(用来算最高分，最低分，平均分)
        Map<String, UsualScoreOverviewDto> usualScoreOverviewDtoMap = new LinkedHashMap<>();//整体平常成绩分数图
//        Map<Long, Map<String, List<Double>>> finalScoreTargetMap = new LinkedHashMap<>();//汇总目标分数(用来算最高分，最低分，平均分)
        Map<Long, List<Double>> finalScoreTargetMap = new LinkedHashMap<>();//汇总目标分数(用来算最高分，最低分，平均分)
        Map<Long, Map<String, List<Double>>> finalScoreDimensionMap = new LinkedHashMap<>();//汇总知识点分数(用来算最高分，最低分，平均分)
//        LinkedMultiValueMap<Long, UsualScoreOverviewDto> oList = new LinkedMultiValueMap<>();
        if (!CollectionUtils.isEmpty(finalScoreResultList)) {
            //期末成绩明细map
            Map<String, PaperStructResult> paperStructResultMap = paperStructResultList.stream().collect(Collectors.toMap(k -> k.getMainNumber() + "-" + k.getSubNumber(), Function.identity(), (dto1, dto2) -> dto1));

            List<ReportExamStudentDto> examStudentList = new ArrayList<>(finalScoreResultList.size());
            for (FinalScoreResult finalScoreResult : finalScoreResultList) {
                //解析期末成绩知识点
                JSONArray finalScoreDetailJsonArray = JSONArray.parseArray(finalScoreResult.getFinalScoreDetail());
                for (int i = 0; i < finalScoreDetailJsonArray.size(); i++) {
                    JSONObject jsonObject = finalScoreDetailJsonArray.getJSONObject(i);
                    PaperStructResult paperStructResult = paperStructResultMap.get(jsonObject.get("name").toString());
                    if (Objects.nonNull(paperStructResult)) {
                        List<CourseTargetDto> courseTargetDtoList = paperStructResult.getTargetList();
                        for (CourseTargetDto courseTargetDto : courseTargetDtoList) {
                            if (!finalScoreTargetMap.containsKey(courseTargetDto.getTargetId())) {

                            } else {

                            }
                            List<DimensionDto> dimensionDtoList = courseTargetDto.getDimensionList();
                        }
                    }
                }

                //解析平时作业分数图
                JSONArray usualScoreJsonArray = JSONArray.parseArray(finalScoreResult.getUsualScore());
                List<ReportExamStudentUsualScoreObjDto> scoreList = new ArrayList<>(usualScoreJsonArray.size());
                for (int i = 0; i < usualScoreJsonArray.size(); i++) {
                    JSONObject jsonObject = usualScoreJsonArray.getJSONObject(i);
                    if (!usualScoreMap.containsKey(jsonObject.get("name"))) {
                        usualScoreMap.put(jsonObject.get("name").toString(), new ArrayList<>(Arrays.asList(Objects.nonNull(jsonObject.get("score")) ? Double.parseDouble(jsonObject.get("score").toString()) : 0d)));
                    } else {
                        List<Double> list = usualScoreMap.get(jsonObject.get("name").toString());
                        list.add(Objects.nonNull(jsonObject.get("score")) ? Double.parseDouble(jsonObject.get("score").toString()) : 0d);
                        usualScoreMap.put(jsonObject.get("name").toString(), list);
                    }
                    scoreList.add(new ReportExamStudentUsualScoreObjDto(jsonObject.get("name").toString(), Objects.nonNull(jsonObject.get("score")) ? Double.parseDouble(jsonObject.get("score").toString()) : 0d));
                }
                finalScoreResult.setUsualScoreDto(new ReportExamStudentUsualScoreDto(scoreList));
                examStudentList.add(new ReportExamStudentDto(finalScoreResult));
            }
        }

        //计算平常成绩最高分，最低分，平均分
        usualScoreMap.forEach((k, v) -> {
            Double maxScore = v.stream().mapToDouble(Double::doubleValue).max().orElse(0d);
            Double minScore = v.stream().mapToDouble(Double::doubleValue).min().orElse(0d);
            Double avgScore = new BigDecimal(v.stream().mapToDouble(Double::doubleValue).average().orElse(0d)).setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();
            usualScoreOverviewDtoMap.put(k, new UsualScoreOverviewDto(k, maxScore, minScore, avgScore));
        });

        //TODO 课程考核成绩评价结果-待曹子轩补充目标/评价方式数据，有测试数据
        ReportCourseEvaluationResultDto reportCourseEvaluationResultDto = null;
        if (!CollectionUtils.isEmpty(paperStructResultList)) {
            Map<Long, String> targetMap = new LinkedHashMap<>();
            for (PaperStructResult paperStructResult : paperStructResultList) {
                List<CourseTargetDto> courseTargetDtoList = paperStructResult.getTargetList();
                for (CourseTargetDto courseTargetDto : courseTargetDtoList) {
                    targetMap.put(courseTargetDto.getTargetId(), courseTargetDto.getTargetName());
                }
            }
            List<CourseTargetDto> targetList = new ArrayList<>(targetMap.size());
            AtomicInteger sequenceId = new AtomicInteger(1);//流程连接id
            AtomicReference<Double> targetValue = new AtomicReference<>(0.0d);
            targetMap.forEach((k, v) -> {
                List<ReportEvaluationDto> reportEvaluationDtoList = new ArrayList<>();
                reportEvaluationDtoList.add(new ReportEvaluationDto("平时作业" + sequenceId.get(), "30%", 40d, 25d));
                reportEvaluationDtoList.add(new ReportEvaluationDto("期末考试", "70%", 60d, 40d));
                CourseTargetDto courseTargetDto = new CourseTargetDto(k, v, Long.parseLong(String.valueOf(sequenceId.get())), "评价方式" + sequenceId.get(), "评价方式描述" + sequenceId.get(), reportEvaluationDtoList, 0.5d + sequenceId.get());
                targetList.add(courseTargetDto);
                targetValue.set(targetValue.get().doubleValue() + courseTargetDto.getEvaluationValue().doubleValue());
                sequenceId.incrementAndGet();
            });
            reportCourseEvaluationResultDto = new ReportCourseEvaluationResultDto("测试课程总目标1", targetValue.get(), targetList);
            trBasicInfo.setCourseEvaluationResult(JacksonUtil.parseJson(reportCourseEvaluationResultDto));
        }

        //课程目标达成评价明细结果-考生集合


        //课程目标达成评价明细结果-课程目标达成评价值图
        trBasicInfoService.saveOrUpdate(trBasicInfo);

        trBasicInfo.setReportResult(new ReportResult(reportCommonDto, reportCourseBasicInfoDto, reportCourseEvaluationSpreadDto, reportCourseEvaluationResultDto, null));
        return ResultUtil.ok(trBasicInfo.getReportResult());
    }

    @ApiOperation(value = "保存报告")
    @RequestMapping(value = "/report/save", method = RequestMethod.POST)
    @OperationLogDetail(customizedOperationType = CustomizedOperationTypeEnum.UPDATE)
    @ApiResponses({@ApiResponse(code = 200, message = "保存报告", response = Object.class)})
    public Object reportSave(@ApiParam(value = "保存报告结构", required = true) @Valid @RequestBody TRBasicInfo trBasicInfo, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultUtil.error(bindingResult.getAllErrors().get(0).getDefaultMessage());
        }
        SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
        TRBasicInfo trBasicInfoDb = trBasicInfoService.getOne(new QueryWrapper<TRBasicInfo>().lambda().eq(TRBasicInfo::getExamId, trBasicInfo.getExamId()).eq(TRBasicInfo::getCourseCode, trBasicInfo.getCourseCode()).eq(TRBasicInfo::getPaperNumber, trBasicInfo.getPaperNumber()));
        if (Objects.isNull(trBasicInfoDb)) {
            trBasicInfoDb = new TRBasicInfo(trBasicInfoDb, sysUser.getId());
        } else {
            trBasicInfoDb.updateInfo(trBasicInfo, sysUser.getId());
        }
        return ResultUtil.ok(trBasicInfoService.saveOrUpdate(trBasicInfo));
    }

    @ApiOperation(value = "导出报告")
    @RequestMapping(value = "/report/export", method = RequestMethod.POST)
    @OperationLogDetail(customizedOperationType = CustomizedOperationTypeEnum.EXPORT)
    @ApiResponses({@ApiResponse(code = 200, message = "下载成功", response = Object.class)})
    public void reportExport(@ApiParam(value = "考试id", required = true) @RequestParam Long examId,
                             @ApiParam(value = "科目编码", required = true) @RequestParam String courseCode,
                             @ApiParam(value = "试卷编号", required = true) @RequestParam String paperNumber) throws IOException {
        TRBasicInfo trBasicInfo = trBasicInfoService.getOne(new QueryWrapper<TRBasicInfo>().lambda().eq(TRBasicInfo::getExamId, examId).eq(TRBasicInfo::getCourseCode, courseCode).eq(TRBasicInfo::getPaperNumber, paperNumber));
        Objects.requireNonNull(trBasicInfo, "没有报告信息");
    }

    /**
     * 计算分数范围
     *
     * @param finalScoreDto
     * @param examId
     * @param paperNumber
     * @param markPaper
     */
    protected void fillScoreRange(FinalScoreDto finalScoreDto, Long examId, String paperNumber, MarkPaper markPaper) {
        //分数区间
        Double passScore = markPaper.getPassScore();//及格分
        Objects.requireNonNull(passScore, "及格分为空");

        Double totalScore = markPaper.getTotalScore();
        Objects.requireNonNull(totalScore, "试卷满分为空");

        List<ScoreRangeVo> scoreRange = new ArrayList<>();
        finalScoreDto.setScoreRange(scoreRange);

        //开始、结束分数区间
        Double startInterval = 0.5d, endInterval = 9.5d, increaseStartScore = passScore - startInterval,
                increaseEndScore = increaseStartScore + startInterval + endInterval;
        scoreRange.add(getScoreRangeVo(finalScoreDto.getStudentCount(), examId, markPaper.getCourseCode(), paperNumber, 1.0d, increaseStartScore));
        increaseStartScore = increaseStartScore + startInterval;
        increaseEndScore = increaseStartScore + endInterval;
        if (totalScore.doubleValue() - increaseEndScore.doubleValue() == startInterval || increaseEndScore.doubleValue() >= totalScore.doubleValue()) {
            increaseEndScore = totalScore;
            scoreRange.add(getScoreRangeVo(finalScoreDto.getStudentCount(), examId, markPaper.getCourseCode(), paperNumber, increaseStartScore, increaseEndScore));
        } else {
            while (increaseEndScore.doubleValue() < totalScore.doubleValue()) {
                scoreRange.add(getScoreRangeVo(finalScoreDto.getStudentCount(), examId, markPaper.getCourseCode(), paperNumber, increaseStartScore, increaseEndScore));
                increaseStartScore = increaseEndScore + startInterval;
                increaseEndScore = increaseStartScore + endInterval;
                if (totalScore.doubleValue() - increaseEndScore.doubleValue() == startInterval || increaseEndScore.doubleValue() >= totalScore.doubleValue()) {
                    increaseEndScore = totalScore;
                    scoreRange.add(getScoreRangeVo(finalScoreDto.getStudentCount(), examId, markPaper.getCourseCode(), paperNumber, increaseStartScore, increaseEndScore));
                }
            }
        }
    }

    /**
     * 计算分数比率
     *
     * @param toltal
     * @param examId
     * @param courseCode
     * @param paperNumber
     * @param start
     * @param end
     * @return
     */
    protected ScoreRangeVo getScoreRangeVo(int toltal, Long examId, String courseCode, String paperNumber, Double start, Double end) {
        int count = tcFinalScoreService.getCountByScoreRange(examId, courseCode, paperNumber, start, end);
        Double rate = null;
        if (toltal != 0) {
            rate = Calculator.multiply(count, toltal, 2);
        }
        return new ScoreRangeVo(count, start, end, rate);
    }
}
